# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import json
import logging
import subprocess
from functools import cached_property
from pathlib import Path

from vasp_manager.utils import LoggerAdapter, change_directory

logger = logging.getLogger(__name__)


class JobManager:
    """
    Handles job submission and status monitoring
    """

    def __init__(
        self,
        calc_dir,
        exe_name="vasp.q",
        jobid_name="jobid",
        config_dir=None,
        manager_name=None,
        ignore_personal_errors=True,
    ):
        """
        Args:
            calc_dir (str | Path): base directory of job
            exe_name (str): filename for slurm job submission
            jobid_name (str): filename for storing slurm jobid
            config_dir (str | Path): path to directory containing configuration files
            manager_name (str): name for logging purposes
            ignore_personal_errors (bool): if True, ignore job submission errors
                if on personal computer
        """
        self.calc_dir = Path(calc_dir)
        self.config_dir = Path(config_dir) if config_dir else self.calc_dir.parents[1]
        self.ignore_personal_errors = ignore_personal_errors
        self.exe_name = exe_name
        self.jobid_name = jobid_name
        self.manager_name = manager_name if manager_name else str(self.calc_dir)
        self._jobid = None
        self._job_complete = None
        self.logger = LoggerAdapter(logging.getLogger(__name__), self.manager_name)

    @property
    def computing_config_dict(self):
        fname = "computing_config.json"
        fpath = self.config_dir / fname
        if fpath.exists():
            with open(fpath) as fr:
                computing_config = json.load(fr)
        else:
            raise Exception(f"No {fname} found in {self.config_dir.absolute()}")
        return computing_config

    @cached_property
    def computer(self):
        return self.computing_config_dict["computer"]

    @cached_property
    def user_id(self):
        return self.computing_config_dict[self.computer]["user_id"]

    @cached_property
    def mode(self):
        return self.calc_dir.name

    @property
    def job_exists(self):
        jobid_path = self.calc_dir / self.jobid_name
        if not jobid_path.exists():
            return False
        if jobid_path.stat().st_size == 0:
            return False

        with open(jobid_path) as fr:
            jobid = fr.read().strip()
        self.jobid = jobid
        return True

    @property
    def jobid(self):
        if not self.job_exists:
            raise Exception(f"jobid has not been set in {self.calc_dir}")
        return self._jobid

    @jobid.setter
    def jobid(self, job_value):
        # some criteria to make sure jobid actually looks reasonable?
        # but I think SLURM will throw and error if sbatch fails
        try:
            jobid_int = int(job_value)
        except Exception as e:
            raise RuntimeError(f"Tried to set jobid={job_value}\n{e}")
        self._jobid = jobid_int

    def submit_job(self):
        """
        Submits job, making sure to not make duplicate jobs
        """
        if self.job_exists:
            self.logger.info(f"{self.mode.upper()} job already exists")
            return True

        if "personal" in self.computer:
            msg = (
                f"Cannot submit {self.mode.upper()} job for on personal computer\n"
                "\tIgnoring job submission..."
            )
            self.logger.debug(msg)
            return True

        qpath = self.calc_dir / self.exe_name
        if not qpath.exists():
            self.logger.info(f"No {self.exe_name} file in {self.calc_dir}")
            # return False here instead of catching an exception
            # This enables job resubmission by letting the calling function
            # know that the calculation needs to be restarted
            return False

        submission_call = f"sbatch {self.exe_name}"
        with change_directory(self.calc_dir):
            submission_call_output = (
                subprocess.check_output(submission_call, shell=True)
                .decode("utf-8")
                .strip()
            )
            jobid = submission_call_output.split(" ")[3]
            self.jobid = jobid
            with open(self.jobid_name, "w+") as fw:
                fw.write(f"{jobid}\n")
        self.logger.info(f"Submitted job {jobid}")
        return True

    @property
    def job_complete(self):
        if self._job_complete is None and self.job_exists:
            self._job_complete = self._check_job_complete()
        return self._job_complete

    def _check_job_complete(self):
        """Returns True if job done"""
        if self.computer == "personal":
            msg = "Cannot check job on personal computer\n\tIgnoring job status check..."
            self.logger.debug(msg)
            # This enables job resubmission by letting the calling function
            # continue anyways
            return True

        check_queue_call = f"squeue -u {self.user_id}"
        queue_call = (
            subprocess.check_output(check_queue_call, shell=True)
            .decode("utf-8")
            .splitlines()
        )
        for line in queue_call:
            line = line.strip().split()
            if str(self.jobid) in line:
                return False
        return True



"""
vasp_manager

A python package to run and analyze VASP calculations
"""

__package__ = "vasp_manager"
__version__ = "1.3.2"
__author__ = "Dale Gaines II"

import logging

from vasp_manager.vasp_manager import VaspManager

logging.getLogger(__name__).addHandler(logging.NullHandler())



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

from functools import cached_property
from pathlib import Path

import numpy as np
from pymatgen.core import Structure

from vasp_manager.utils import pgrep


class ElasticAnalyzer:
    def __init__(
        self,
        cij,
        structure,
        rounding_precision=3,
    ):
        """
        Args:
            cij (6x6 array-like[float]): stiffness tensor in Voigt notation in
                units of GPa
                -- careful! VASP does not output stiffness tensors in this notation
            structure (pmg.Structure)
            rounding_precision (int): precision to round calculated quantities
        """
        self.cij = cij
        self.structure = structure
        self.rounding_precision = rounding_precision
        self._results = None

    @property
    def cij(self):
        return self._cij

    @cij.setter
    def cij(self, values):
        values = np.asarray(values)
        if values.shape != (6, 6):
            raise ValueError(
                "cij must be a 6x6 array," f"found shape={values.shape} instead"
            )
        if values.dtype != float:
            raise TypeError(
                "cij must be a 6x6 array of floats," f" found type={values.dtype} instead"
            )
        self._cij = values

    @property
    def structure(self):
        return self._structure

    @structure.setter
    def structure(self, value):
        if not isinstance(value, Structure):
            raise TypeError(
                "Structure must be pymatgen Structure object,"
                f" found type={type(value)} instead"
            )
        self._structure = value

    @property
    def density(self):
        return self.structure.density

    @property
    def rounding_precision(self):
        return self._rounding_precision

    @rounding_precision.setter
    def rounding_precision(self, value):
        if not isinstance(value, int):
            raise TypeError(
                f"Could not set rounding_precision to {value}. rounding_precision must"
                " be an int, found type={type(value)} instead)"
            )
        self._rounding_precision = value

    @staticmethod
    def change_elastic_constants_from_vasp(vasp_elastic_tensor):
        """
        VASP ordering of elastic constants is not the same as those expected
        by my equations below
        We should expect from Voigt notation:
        1: 11 or xx
        2: 22 or yy
        3: 33 or zz
        4: 23 or yz
        5: 13 or xz
        6: 12 or xy
        but VASP differs as it presents 1, 2, 3, 6 (xy), 4 (yz), 5 (xz)
        This function performs swapping to match expectations

        Args:
            vasp_elastic_tensor (6x6 np.array[float])
        Returns:
            elastic_tensor (6x6 np.array[float]): reordered to match Voigt notation
        """
        elastic_tensor = np.asarray(vasp_elastic_tensor).copy()
        for j in range(6):
            elastic_tensor[3, j], elastic_tensor[4, j], elastic_tensor[5, j] = (
                elastic_tensor[4, j],
                elastic_tensor[5, j],
                elastic_tensor[3, j],
            )
        for i in range(6):
            elastic_tensor[i, 3], elastic_tensor[i, 4], elastic_tensor[i, 5] = (
                elastic_tensor[i, 4],
                elastic_tensor[i, 5],
                elastic_tensor[i, 3],
            )
        return elastic_tensor

    @staticmethod
    def get_compliance_tensor(cij):
        """
        Args:
            cij (6x6 np.arrray[float]): stiffness tensor
        Returns:
            sij (6x6 np.arrray[float]): compliance tensor
        """
        sij = np.linalg.inv(cij)
        return sij

    @staticmethod
    def get_VRH_average(mod1, mod2):
        """
        Args:
            mod1 (float):  B or G Voigt
            mod2 (float): B or G Reuss
        Returns:
            VRH_average (float)
        """
        return (mod1 + mod2) / 2.0

    @staticmethod
    def get_B_Reuss(sij):
        """
        Args:
            sij (6x6 np.arrray[float]): compliance tensor
        Returns:
            B_Reuss (float): Reuss bulk modulus
        """
        B_Reuss = 1 / (
            (sij[0, 0] + sij[1, 1] + sij[2, 2]) + 2 * (sij[0, 1] + sij[1, 2] + sij[2, 0])
        )
        return B_Reuss

    @staticmethod
    def get_B_Voigt(cij):
        """
        Args:
            cij (6x6 np.arrray[float]): compliance_tensor
        Returns:
            B_Reuss (float): Reuss bulk modulus
        """
        B_Voigt = (
            (cij[0, 0] + cij[1, 1] + cij[2, 2]) + 2 * (cij[0, 1] + cij[1, 2] + cij[2, 0])
        ) / 9
        return B_Voigt

    @staticmethod
    def get_G_Reuss(sij):
        """
        Args:
            sij (6x6 np.arrray[float]): compliance tensor
        Returns:
            B_Reuss (float): Reuss bulk modulus
        """
        G_Reuss = 15 / (
            4 * (sij[0, 0] + sij[1, 1] + sij[2, 2])
            - 4 * (sij[0, 1] + sij[1, 2] + sij[2, 0])
            + 3 * (sij[3, 3] + sij[4, 4] + sij[5, 5])
        )
        return G_Reuss

    @staticmethod
    def get_G_Voigt(cij):
        """
        Args:
            cij (6x6 np.arrray[float]): stiffness tensor
        Returns:
            G_Reuss (float): Reuss shear modulus
        """
        G_Voigt = (
            (cij[0, 0] + cij[1, 1] + cij[2, 2])
            - (cij[0, 1] + cij[1, 2] + cij[2, 0])
            + 3 * (cij[3, 3] + cij[4, 4] + cij[5, 5])
        ) / 15
        return G_Voigt

    @staticmethod
    def get_vl(bulkmod, shearmod, density):
        vl = np.sqrt((bulkmod + 4 / 3 * shearmod) / density)
        return vl

    @staticmethod
    def get_vt(shearmod, density):
        vt = np.sqrt(shearmod / density)
        return vt

    @staticmethod
    def get_vs(bulkmod, shearmod, density):
        vl = np.sqrt((bulkmod + 4 / 3 * shearmod) / density)
        vt = np.sqrt(shearmod / density)
        vs = (1 / 3 * (1 / vl**3 + 2 / vt**3)) ** (-1 / 3)
        return vs

    @staticmethod
    def check_elastically_unstable(cij):
        "returns True if compound is elastically unstable"
        eigenvalues, eigenvectors = np.linalg.eig(cij)
        born_critera_satisfied = np.all(eigenvalues > 0)
        elastically_unstable = not born_critera_satisfied
        return elastically_unstable

    @cached_property
    def sij(self):
        # DO NOT ROUND
        return self.get_compliance_tensor(self.cij)

    @cached_property
    def b_reuss(self):
        return np.round(self.get_B_Reuss(self.sij), self.rounding_precision)

    @cached_property
    def b_voigt(self):
        return np.round(self.get_B_Voigt(self.cij), self.rounding_precision)

    @cached_property
    def b_vrh(self):
        return np.round(
            self.get_VRH_average(self.b_reuss, self.b_voigt), self.rounding_precision
        )

    @cached_property
    def g_reuss(self):
        return np.round(self.get_G_Reuss(self.sij), self.rounding_precision)

    @cached_property
    def g_voigt(self):
        return np.round(self.get_G_Voigt(self.cij), self.rounding_precision)

    @cached_property
    def g_vrh(self):
        return np.round(
            self.get_VRH_average(self.g_reuss, self.g_voigt), self.rounding_precision
        )

    @cached_property
    def elastically_unstable(self):
        return self.check_elastically_unstable(self.cij)

    @cached_property
    def vt(self):
        return np.round(self.get_vt(self.g_vrh, self.density), self.rounding_precision)

    @cached_property
    def vl(self):
        return np.round(
            self.get_vl(self.b_vrh, self.g_vrh, self.density), self.rounding_precision
        )

    @cached_property
    def vs(self):
        return np.round(
            self.get_vs(self.b_vrh, self.g_vrh, self.density), self.rounding_precision
        )

    def _analyze_elastic(self, properties=None):
        """
        Grabs important quantities from the elastic calculation results

        Args:
            properties (list): names of properties to calculate
        Returns:
            elastic_dict (dict): dict of extracted info from
                elastic constants
        """
        properties_map = {
            "B_Reuss": "b_reuss",
            "B_Voigt": "b_voigt",
            "B_VRH": "b_vrh",
            "G_Reuss": "g_reuss",
            "G_Voigt": "g_voigt",
            "G_VRH": "g_vrh",
            "unstable": "elastically_unstable",
            "elastic_tensor": "cij",
            "vl": "vl",
            "vt": "vt",
            "vs": "vs",
        }
        if properties is None:
            properties = list(properties_map.keys())

        elastic_dict = {}
        for property in properties:
            elastic_dict[property] = getattr(self, properties_map[property])
        return elastic_dict

    @property
    def results(self):
        if self._results is None:
            self._results = self._analyze_elastic()
        return self._results

    @classmethod
    def from_calc_dir(cls, calc_dir, **kwargs):
        """
        Method to construct ElasticAnalyzer from a calculation directory
        The directory should contain a POSCAR and an OUTCAR with elastic constants

        Args:
            calc_dir (str | Path)
        """
        calc_dir = Path(calc_dir)
        if not calc_dir.exists():
            raise ValueError(f"Could not set calc_dir to {calc_dir} as it does not exist")

        structure = Structure.from_file(calc_dir / "POSCAR")
        outcar_glob = list(calc_dir.glob("OUTCAR*"))
        if len(outcar_glob) == 0:
            raise FileNotFoundError(f"No OUTCAR available in {calc_dir}")
        outcar_filepath = outcar_glob[0]

        elastic_filepath = calc_dir / "elastic_constants.txt"
        if not elastic_filepath.exists():
            # Scrapes OUTCAR for elastic constants and write to elastic_constants.txt
            elastic_table = pgrep(
                outcar_filepath,
                str_to_grep="TOTAL ELASTIC MOD",
                stop_after_first_match=True,
                after=8,
                as_string=True,
            )
            with open(elastic_filepath, "w+") as fw:
                fw.write(elastic_table)

        # Read the elastic constants and construct the stiffness tensor
        with open(elastic_filepath) as fr:
            raw_elastic_data = fr.readlines()
        try:
            # Skip first 3 rows as they are just header
            # Skip the first column as it contains xx, xy, etc
            elastic_data = [line.strip().split()[1:] for line in raw_elastic_data[3:]]
            # Divide by 10 to get GPa instead of kBar
            elastic_tensor = np.array(elastic_data, dtype=float) / 10.0
            # VASP only outputs the stiffness tensor to 4 places
            elastic_tensor = np.round(elastic_tensor, 4)
            elastic_tensor = ElasticAnalyzer.change_elastic_constants_from_vasp(
                elastic_tensor
            )
        except Exception as e:
            raise RuntimeError(
                "Could not construct elastic tensor from OUTCAR. Please verify the"
                " OUTCAR contains a full elastic tensor.\n"
                f"{e}"
            )
        return cls(cij=elastic_tensor, structure=structure, **kwargs)



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import gzip
import json
import logging
import os
from collections import deque
from contextlib import contextmanager
from pathlib import Path

import numpy as np
from pymatgen.core import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer


@contextmanager
def change_directory(new_dir):
    prev_dir = os.getcwd()
    os.chdir(os.path.expanduser(new_dir))
    try:
        yield
    finally:
        os.chdir(prev_dir)


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types"""

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class LoggerAdapter(logging.LoggerAdapter):
    def __init__(self, logger, prefix, separator=" -- "):
        super(LoggerAdapter, self).__init__(logger, {})
        self.prefix = prefix
        self.separator = separator

    def process(self, msg, kwargs):
        return f"{self.prefix}{self.separator}{msg}", kwargs


def get_pmg_structure_from_poscar(
    poscar_path,
    to_process=True,
    primitive=True,
    symprec=1e-03,
    angle_tolerance=-1.0,
    international_monoclinic=False,
    return_spacegroup=False,
):
    """
    Args:
        poscar_path (str | Path)
        to_process (bool): if True, get standard reduced structure
        primitive (bool): if True, get primitive structure, else get
            conventional structure
        symprec (float): symprec for SpacegroupAnalyzer
        angle_tolerance (float): angle tolerance for SpacegroupAnalyzer
        international_monoclinic (bool): if True, convert to proper
            international convention such that beta is the non-right angle.
            WARNING: setting True is not compatible with pymatgen kpaths
        return_spacegroup (bool): if True, return spacegroup number
    Returns:
        structure (pmg.Structure): structure from POSCAR
    """
    structure = Structure.from_file(poscar_path)
    if to_process:
        sga = SpacegroupAnalyzer(
            structure, symprec=symprec, angle_tolerance=angle_tolerance
        )
        if primitive:
            structure = sga.get_primitive_standard_structure(
                international_monoclinic=international_monoclinic
            )
        else:
            structure = sga.get_conventional_standard_structure(
                international_monoclinic=international_monoclinic
            )
        if return_spacegroup:
            spacegroup = sga.get_space_group_number()
            return structure, spacegroup
    return structure


def pcat(file_names):
    """
    Custom python-only replacement for cat

    Args:
        file_names (list): names of files to cat together
    Returns:
        catted (str)
    """
    file_contents = []
    if isinstance(file_names, (str, Path)):
        file_names = [file_names]
    for file_name in file_names:
        with open(file_name) as fr:
            file_content = fr.read()
        file_contents.append(file_content)
    catted = "".join(file_content for file_content in file_contents)
    return catted


def pgrep(
    file_name,
    str_to_grep,
    stop_after_first_match=False,
    after=None,
    as_string=False,
):
    """
    Custom python-only replacement for grep

    Args:
        file_name (str | Path): path of file
        str_to_grep (str): target string
        stop_after_first_match (bool): if True, stop after first found instance of
            str_to_grep
        after (int): if not None, return {after} lines found after str_to_grep
        as_str (bool): if as_string, return a single string, else return splitlines
    Returns:
        matches (str | list)
    """
    opener = gzip.open if ".gz" in str(file_name) else open
    matches = []
    line_idx_to_include = set()
    with opener(file_name, "rt") as fr:
        for line_idx, line in enumerate(fr):
            if str_to_grep in line:
                matches.append(line.strip("\n"))
                if after is not None:
                    line_idx_to_include.update(range(line_idx + 1, line_idx + after + 1))
            if line_idx in line_idx_to_include:
                matches.append(line.strip("\n"))
                line_idx_to_include.discard(line_idx)
            if stop_after_first_match:
                if len(matches) != 0 and len(line_idx_to_include) == 0:
                    break
    if as_string:
        matches = "\n".join([line for line in matches])
    return matches


def phead(file_name, n_head=1, as_string=False):
    """
    Custom python-only replacement for head

    Args:
        file_name (str): path of file
        n_head (int): n lines to head
        as_str (bool): if as_string, return a single string, else return splitlines
    Returns:
        head (str | list)
    """
    opener = gzip.open if ".gz" in str(file_name) else open
    head = []
    with opener(file_name, "rt") as fr:
        for i, line in enumerate(fr):
            if i < n_head:
                head.append(line.strip("\n"))
    if as_string:
        head = "\n".join([line for line in head])
    return head


def ptail(file_name, n_tail=1, as_string=False):
    """
    Custom python-only replacement for grep

    Args:
        file_name (str): path of file
        n_tail (int): n lines to tail
        as_str (bool): if as_string, return a single string, else return splitlines
    Returns:
        tail (str | list)
    """
    opener = gzip.open if ".gz" in str(file_name) else open
    tail = deque(maxlen=n_tail)
    with opener(file_name, "rt") as fr:
        for line in fr:
            tail.append(line.strip("\n"))
    tail = list(tail)
    if as_string:
        tail = "\n".join([line for line in tail])
    return tail


def make_potcar_anonymous(input_file_name, output_file_name=None):
    """
    Replace full POTCAR with only single POTCAR names

    Args:
        input_file_name (str | Path): path of POTCAR file
        output_file_name (str | Path | None): path to write anonymized POTCAR
            if None, write to the location of input_f_name
    Returns:
        None
    """
    if output_file_name is None:
        output_file_name = input_file_name

    with open(input_file_name, "rt") as fr:
        full_potcar_text = [line.strip() for line in fr.readlines()]
    trimmed_potcar_lines = []
    for line in full_potcar_text:
        if "TITEL" in line:
            trimmed_potcar_lines.append(line.split("=")[1].strip())
    trimmed_potcar_string = "\n".join([line for line in trimmed_potcar_lines])
    with open(output_file_name, "w+") as fw:
        fw.write(trimmed_potcar_string)



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import json
import logging
from functools import cached_property
from importlib.metadata import version
from multiprocessing import Pool
from pathlib import Path

import numpy as np
from tqdm import tqdm

from vasp_manager.calculation_manager import (
    BulkmodCalculationManager,
    ElasticCalculationManager,
    RlxCalculationManager,
    RlxCoarseCalculationManager,
    StaticCalculationManager,
)
from vasp_manager.utils import NumpyEncoder

logger = logging.getLogger(__name__)

ASCII_LOGO = r"""
 __      __             __  __
 \ \    / /            |  \/  |
  \ \  / /_ _ ___ _ __ | \  / | __ _ _ __   __ _  __ _  ___ _ __
   \ \/ / _` / __| '_ \| |\/| |/ _` | '_ \ / _` |/ _` |/ _ \ '__|
    \  / (_| \__ \ |_) | |  | | (_| | | | | (_| | (_| |  __/ |
     \/ \__,_|___/ .__/|_|  |_|\__,_|_| |_|\__,_|\__, |\___|_|
                 | |                              __/ |
                 |_|                             |___/ v{}
""".format(
    version("vasp_manager")
)


class VaspManager:
    """
    Handles set up and execution of each CalculationManager
        (rlx-coarse, rlx, static, bulkmod, elastic)
    """

    def __init__(
        self,
        calculation_types,
        material_dirs=None,
        to_rerun=True,
        to_submit=True,
        ignore_personal_errrors=True,
        tail=5,
        use_multiprocessing=False,
        ncore=None,
        calculation_manager_kwargs={},
        max_reruns=3,
        magmom_per_atom_cutoff=0.0,
        sort_by=str,
    ):
        """
        Args:
            calculation_types (list[str]): list of calculation types
            material_dirs (list[str] | str): list of material directory paths OR
                name of calculations directory
            to_rerun (bool): if True, rerun failed calculations
            to_submit (bool): if True, submit calculations
            ignore_personal_errors (bool): if True, ignore job submission errors
                if on personal computer
            tail (int): number of last lines from stdout.txt to log in debugging
                if job failed
            use_multiprocessing (bool): if True, use pool.map()
            ncore (int): if ncore, use {ncore} for multiprocessing
                if None, defaults to minimum(number of materials, 4)
            calculation_manager_kwargs (dict): contains subdictionaries for each
                calculation type. Each subdictorary can be filled with extra kwargs
                to pass to its associated CalculationManager during instantiation
            max_reruns (int): the maximum number of times a rlx-coarse or rlx
                calculation can run before refusing to continue
                Note: other modes don't make archives, so they are not affected
                by this
            magmom_per_atom_cutoff (float): calculations that result in
                magmom_per_atom less than this parameter will be automatically
                rerun without spin-polarization
            sort_by (callable): function to sort the keys of the result dictionary
        """
        print(ASCII_LOGO)
        self.sort_by = sort_by
        self.calculation_types = calculation_types
        self.material_dirs = material_dirs
        self.to_rerun = to_rerun
        self.to_submit = to_submit
        self.ignore_personal_errors = ignore_personal_errrors
        self.tail = tail
        self.use_multiprocessing = use_multiprocessing
        self.ncore = ncore
        self.calculation_manager_kwargs = calculation_manager_kwargs
        self.max_reruns = max_reruns
        self.magmom_per_atom_cutoff = magmom_per_atom_cutoff

        self.calculation_managers = self._get_all_calculation_managers()
        # self.base_dir is set in material_dirs.setter
        self.results_path = self.base_dir / "results.json"
        self.results = None

    @property
    def calculation_types(self):
        return self._calculation_types

    @calculation_types.setter
    def calculation_types(self, values):
        if not isinstance(values, list):
            raise TypeError("calculation_types must be a list")
        self._calculation_types = values

    @property
    def ncore(self):
        return self._ncore

    @ncore.setter
    def ncore(self, value):
        if value is None:
            value = int(min(len(self.material_dirs), 4))
            if self.use_multiprocessing:
                print(
                    "WARNING: setting default ncore for multiprocessing to "
                    + f"{value}\n"
                    + "    We strongly recommend you set ncore to the number of "
                    + "available cores."
                )
        if not isinstance(value, int):
            raise Exception
        self._ncore = value

    @property
    def calculation_manager_kwargs(self):
        return self._calculation_manager_kwargs

    @calculation_manager_kwargs.setter
    def calculation_manager_kwargs(self, values):
        if not isinstance(values, dict):
            raise TypeError("calculation_manager_kwargs must be a dictionary")

        supported_kwargs = ["primitive", "from_scratch", "strains"]
        for calc_type in self.calculation_types:
            if calc_type not in values:
                values[calc_type] = {}
            else:
                for kwarg in values[calc_type]:
                    if kwarg not in supported_kwargs:
                        raise ValueError(
                            f"kwarg={kwarg} is not supported for mode={calc_type}"
                        )
        self._calculation_manager_kwargs = values

    @property
    def material_dirs(self):
        return self._material_dirs

    @material_dirs.setter
    def material_dirs(self, values):
        """
        Sets paths for all materials

        Args:
            values (list[str | Path] | str | Path): list of material directory
                paths OR name of calculations dir

                if is list, use that list directly
                if is string, find folders inside of that directory
        """
        match values:
            case str():
                self.base_dir = values
                material_dirs = [d for d in Path(values).glob("*") if d.is_dir()]
            case list() | np.array():
                values = [Path(p) for p in values]
                base_dir = values[0].parent
                for mat_dir in values:
                    if not mat_dir.parent == base_dir:
                        raise Exception("All material_dirs must be in the same directory")
                self.base_dir = base_dir
                material_dirs = values
            case _:
                raise TypeError(
                    "material_dirs must be a directory name or a list of directories"
                )
        # Sort the paths by name
        self._material_dirs = sorted(material_dirs, key=lambda x: self.sort_by(x.name))

    def _get_material_name_from_path(self, material_dir):
        return material_dir.name

    @cached_property
    def material_names(self):
        return [self._get_material_name_from_path(mpath) for mpath in self.material_dirs]

    @property
    def results(self):
        return self._results

    @results.setter
    def results(self, value):
        if value is None:
            if self.results_path.exists():
                with open(self.results_path, "r") as fr:
                    self._results = json.load(fr)
                for mat_name in self.material_names:
                    if mat_name not in self._results:
                        self._results[mat_name] = {}
                self._results = dict(
                    sorted(self._results.items(), key=lambda kv: self.sort_by(kv[0]))
                )
            else:
                self._results = {mat_name: {} for mat_name in self.material_names}

    def _check_calc_by_result(self, material_name, calc_type):
        """
        Checks if job has been completed and analyzed

        Args:
            material_name (str): name of material to check
            calc_type (str): calculation type to check
        Returns:
            (is_done (bool), is_stopped (bool))
        """
        match calc_type:
            case "rlx-coarse" | "rlx" | "static" | "bulkmod" | "elastic":
                is_done = self.results[material_name][calc_type] not in [
                    None,
                    "STOPPED",
                    "not finished",
                ]
                is_stopped = self.results[material_name][calc_type] == "STOPPED"
            case _:
                raise ValueError("Can't find mode {mode} in result")
        return (is_done, is_stopped)

    def _get_calculation_managers(self, material_dir):
        """
        Gets calculation managers for a single material
        """
        calc_managers = []
        for calc_type in self.calculation_types:
            match calc_type:
                case "rlx-coarse":
                    manager = RlxCoarseCalculationManager(
                        material_dir=material_dir,
                        to_rerun=self.to_rerun,
                        to_submit=self.to_submit,
                        ignore_personal_errors=self.ignore_personal_errors,
                        tail=self.tail,
                        max_reruns=self.max_reruns,
                        **self.calculation_manager_kwargs[calc_type],
                    )
                case "rlx":
                    from_coarse_relax = "rlx-coarse" in self.calculation_types
                    manager = RlxCalculationManager(
                        material_dir=material_dir,
                        to_rerun=self.to_rerun,
                        to_submit=self.to_submit,
                        ignore_personal_errors=self.ignore_personal_errors,
                        from_coarse_relax=from_coarse_relax,
                        tail=self.tail,
                        max_reruns=self.max_reruns,
                        magmom_per_atom_cutoff=self.magmom_per_atom_cutoff,
                        **self.calculation_manager_kwargs[calc_type],
                    )
                case "static":
                    from_relax = "rlx" in self.calculation_types
                    manager = StaticCalculationManager(
                        material_dir=material_dir,
                        to_rerun=self.to_rerun,
                        to_submit=self.to_submit,
                        ignore_personal_errors=self.ignore_personal_errors,
                        from_relax=from_relax,
                        tail=self.tail,
                        **self.calculation_manager_kwargs[calc_type],
                    )
                case "bulkmod":
                    from_relax = "rlx" in self.calculation_types
                    manager = BulkmodCalculationManager(
                        material_dir=material_dir,
                        to_rerun=self.to_rerun,
                        to_submit=self.to_submit,
                        ignore_personal_errors=self.ignore_personal_errors,
                        from_relax=from_relax,
                        **self.calculation_manager_kwargs[calc_type],
                    )
                case "elastic":
                    if "rlx" not in self.calculation_types:
                        msg = (
                            "Cannot perform elastic calculation without mode='rlx'"
                            " first"
                        )
                        raise Exception(msg)
                    manager = ElasticCalculationManager(
                        material_dir=material_dir,
                        to_rerun=self.to_rerun,
                        to_submit=self.to_submit,
                        ignore_personal_errors=self.ignore_personal_errors,
                        tail=self.tail,
                        **self.calculation_manager_kwargs[calc_type],
                    )
                case _:
                    raise Exception(f"Calc type {calc_type} not supported")

            calc_managers.append(manager)
        return calc_managers

    def _get_all_calculation_managers(self):
        """
        Gets calculation managers for all materials
        """
        calc_managers = {}
        for material_name, material_dir in zip(self.material_names, self.material_dirs):
            calc_managers[material_name] = self._get_calculation_managers(material_dir)
        return calc_managers

    def _manage_calculations(self, material_name):
        """
        Runs vasp job workflow for a single material
        """
        material_results = {}
        for calc_manager in self.calculation_managers[material_name]:
            if calc_manager.mode in self.results[material_name].keys():
                calc_is_done, calc_is_stopped = self._check_calc_by_result(
                    material_name, calc_manager.mode
                )
                if calc_is_done and not calc_manager.from_scratch:
                    logger.info(
                        f"{material_name} -- {calc_manager.mode.upper()} Successful"
                    )
                    continue

            if calc_manager.stopped:
                logger.info(f"{material_name} -- {calc_manager.mode.upper()} STOPPED")
                material_results[calc_manager.mode] = "STOPPED"
                break

            if not calc_manager.job_exists:
                logger.info(f"{material_name} -- Setting up {calc_manager.mode.upper()}")
                calc_manager.setup_calc()
                match calc_manager.mode:
                    case "rlx-coarse" | "rlx":
                        break
                    case _:
                        pass

            if not calc_manager.is_done:
                match calc_manager.mode:
                    case "rlx-coarse" | "rlx" | "elastic":
                        # don't check further modes as they rely on rlx-coarse
                        # or rlx to be done
                        # break if elastic not done to avoid analysis
                        break
                    case _:
                        # go ahead and check the other modes as they are
                        # independent of each other
                        pass

            material_results[calc_manager.mode] = calc_manager.results
        return (material_name, material_results)

    def _manage_calculations_wrapper(self):
        if self.use_multiprocessing:
            with Pool(self.ncore) as pool:
                results = pool.map(
                    self._manage_calculations, tqdm(self.material_names), 1
                )
        else:
            results = []
            for i, material_name in enumerate(self.material_names):
                print(f"{i+1}/{len(self.material_names)} -- {material_name}", flush=True)
                results.append(self._manage_calculations(material_name))
                print()
        return results

    def run_calculations(self):
        """
        Runs vasp job workflow for all materials
        """
        results = self._manage_calculations_wrapper()
        for material_name, material_result in results:
            self.results[material_name].update(material_result)

        json_str = json.dumps(self.results, indent=2, cls=NumpyEncoder)
        logger.debug(json_str)
        with open(self.results_path, "w+") as fw:
            fw.write(json_str)
        print(f"Dumped to {self.results_path}")
        return self.results

    def summary(self, as_string=True, print_unfinished=False, print_stopped=True):
        """
        Create a string summary of all calculations

        Args:
            as_string (bool):
                if as_string, return string summary
                else, return dict summary
        """
        if not self.results_path.exists():
            raise ValueError(f"Can't find results in {self.results_path}")
        with open(self.results_path) as fr:
            results = json.load(fr)

        summary_dict = {}
        summary_dict["n_total"] = len(results)
        for calc_type in self.calculation_types:
            summary_dict[calc_type] = {}
            summary_dict[calc_type]["n_finished"] = 0
            summary_dict[calc_type]["finished"] = []
            summary_dict[calc_type]["unfinished"] = []
            summary_dict[calc_type]["stopped"] = []

            for material, mat_results in results.items():
                # need to account for case key doesn't yet exist
                if calc_type not in mat_results:
                    summary_dict[calc_type]["unfinished"].append(material)
                else:
                    is_done, is_stopped = self._check_calc_by_result(material, calc_type)
                    if is_done:
                        summary_dict[calc_type]["n_finished"] += 1
                        summary_dict[calc_type]["finished"].append(material)
                    else:
                        if is_stopped:
                            summary_dict[calc_type]["stopped"].append(material)
                        summary_dict[calc_type]["unfinished"].append(material)

        if as_string:
            n_materials = summary_dict["n_total"]
            summary_str = ""
            summary_str += f"Total Materials = {n_materials}\n"
            summary_str += "-" * 30 + "\n"
            for calc_type in self.calculation_types:
                name = calc_type.upper()
                n_finished = summary_dict[calc_type]["n_finished"]
                summary_str += f"{name: <12}{n_finished}/{n_materials} completed\n"
                if print_unfinished:
                    unfinished = summary_dict[calc_type]["unfinished"]
                    if len(unfinished) != 0:
                        summary_str += (
                            " " * 12 + f"{n_materials - n_finished} not completed\n"
                        )
                        summary_str += f"Unfinished {calc_type.upper()}: {unfinished}\n"
                if print_stopped:
                    stopped = summary_dict[calc_type]["stopped"]
                    if len(stopped) != 0:
                        summary_str += f"\tStopped {calc_type.upper()}: {stopped}\n"
            return summary_str
        else:
            return summary_dict



# Copyright (c) Dale Gaines II
# Distributed under the terms of the MIT LICENSE

import logging
from functools import cached_property

from vasp_manager.calculation_manager.base import BaseCalculationManager
from vasp_manager.utils import LoggerAdapter, pgrep, ptail
from vasp_manager.vasp_input_creator import VaspInputCreator

logger = logging.getLogger(__name__)


class RlxCoarseCalculationManager(BaseCalculationManager):
    """
    Runs coarse relaxation job workflow for a single material
    """

    def __init__(
        self,
        material_dir,
        to_rerun,
        to_submit,
        primitive=True,
        ignore_personal_errors=True,
        from_scratch=False,
        tail=5,
        max_reruns=3,
    ):
        """
        For material_dir, to_rerun, to_submit, ignore_personal_errors, and from_scratch,
        see BaseCalculationManager

        Args:
            tail (int): number of last lines to log in debugging if job failed
        """
        self.tail = tail
        self.max_reruns = max_reruns
        super().__init__(
            material_dir=material_dir,
            to_rerun=to_rerun,
            to_submit=to_submit,
            primitive=primitive,
            ignore_personal_errors=ignore_personal_errors,
            from_scratch=from_scratch,
        )
        self._is_done = None
        self._results = None
        self.logger = LoggerAdapter(logging.getLogger(__name__), self.material_name)

    @cached_property
    def mode(self):
        return "rlx-coarse"

    @cached_property
    def poscar_source_path(self):
        return self.material_dir / "POSCAR"

    @cached_property
    def vasp_input_creator(self):
        return VaspInputCreator(
            self.calc_dir,
            mode=self.mode,
            poscar_source_path=self.poscar_source_path,
            primitive=self.primitive,
            name=self.material_name,
        )

    def setup_calc(
        self,
        increase_nodes_by_factor=1,
        increase_walltime_by_factor=1,
        make_archive=False,
    ):
        """
        Sets up a coarse relaxation
        """
        self.vasp_input_creator.increase_nodes_by_factor = increase_nodes_by_factor
        self.vasp_input_creator.increase_walltime_by_factor = increase_walltime_by_factor

        if make_archive:
            self.vasp_input_creator.make_archive_and_repopulate()
        else:
            self.vasp_input_creator.create()

        if self.to_submit:
            job_submitted = self.submit_job()
            # job status returns True if sucessfully submitted, else False
            if not job_submitted:
                # if job didn't submit, try rerunning setup
                self.setup_calc()

    def check_calc(self):
        """
        Checks if calculation has finished and reached required accuracy

        Returns:
            relaxation_successful (bool): if True, relaxation completed successfully
        """
        if not self.job_complete:
            self.logger.info(f"{self.mode.upper()} not finished")
            return False

        stdout_path = self.calc_dir / "stdout.txt"
        if not stdout_path.exists():
            # calculation never actually ran
            # shouldn't get here unless function was called with submit=False
            # or job was manually cancelled
            self.logger.info(f"{self.mode.upper()} Calculation: No stdout.txt available")
            if self.to_rerun:
                self._cancel_previous_job()
                self.setup_calc()
            return False

        vasp_errors = self._check_vasp_errors()
        if len(vasp_errors) > 0:
            all_errors_addressed = self._address_vasp_errors(vasp_errors)
            if all_errors_addressed:
                if self.to_rerun:
                    self.logger.info(f"Rerunning {self.calc_dir}")
                    self.setup_calc(make_archive=True)
            else:
                msg = (
                    f"{self.mode.upper()} Calculation: "
                    "Couldn't address all VASP Errors\n"
                    f"\tVASP Errors: {vasp_errors}\n"
                    "\tRefusing to continue...\n"
                )
                self.logger.error(msg)
                self.stop()
            return False

        tail_output = ptail(stdout_path, n_tail=self.tail, as_string=True)
        grep_output = pgrep(
            stdout_path, "reached required accuracy", stop_after_first_match=True
        )
        if len(grep_output) == 0:
            archive_dirs = list(self.calc_dir.glob("archive*"))
            if len(archive_dirs) >= self.max_reruns - 1:
                self.logger.warning(
                    "Many archives exist, continuing to force based relaxation..."
                )
                return True

            self.logger.warning(f"{self.mode.upper()} FAILED")
            self.logger.debug(tail_output)
            if self.to_rerun:
                self.logger.info(f"Rerunning {self.calc_dir}")
                # increase nodes as its likely the calculation failed
                self.setup_calc(increase_walltime_by_factor=2, make_archive=True)
            return False

        self.logger.info(f"{self.mode.upper()} Calculation: reached required accuracy")
        self.logger.debug(tail_output)
        return True

    @property
    def is_done(self):
        if self._is_done is None:
            self._is_done = self.check_calc()
        return self._is_done

    @property
    def results(self):
        if not self.is_done:
            if self.stopped:
                return "STOPPED"
            else:
                return "not finished"
        else:
            self._results = "done"
        return self._results



import json
import shutil
from pathlib import Path

import importlib_resources
import numpy as np
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core import Structure

from vasp_manager.utils import (
    NumpyEncoder,
    change_directory,
    make_potcar_anonymous,
    pcat,
    pgrep,
    phead,
    ptail,
)

input_string = """\
DAV:   5    -0.677727844685E+01    0.16454E-04   -0.48875E-05  1776   0.450E-02    0.143E-02
DAV:   6    -0.677727668990E+01    0.17570E-05   -0.21379E-06   872   0.154E-02
   4 F= -.67772767E+01 E0= -.67772767E+01  d E =-.421689E-02
curvature:  -0.59 expect dE=-0.179E-04 dE for cont linesearch -0.179E-04
trial: gam= 0.00000 g(F)=  0.914E-61 g(S)=  0.301E-04 ort =-0.171E-03 (trialstep = 0.280E+01)
search vector abs. value=  0.301E-04
reached required accuracy - stopping structural energy minimisation
PROFILE, used timers:     369"""


def test_change_directory(tmp_path):
    with change_directory(tmp_path):
        assert Path.cwd().name == tmp_path.name


def test_numpy_encoder(tmp_path):
    data = {"array": np.linspace(0, 10, 11)}
    with open(tmp_path / "data.json", "w+") as fw:
        json.dump(data, fw, indent=2, cls=NumpyEncoder)


def test_pmg_structure(tmp_path):
    # original unsymmetrized structure
    orig_poscar_path = importlib_resources.files("vasp_manager").joinpath(
        str(Path("tests") / "calculations" / "material" / "POSCAR")
    )
    # symmetrized structure
    poscar_path = importlib_resources.files("vasp_manager").joinpath(
        str(Path("tests") / "calculations" / "material" / "rlx-coarse" / "POSCAR")
    )
    orig_poscar_structure = Structure.from_file(orig_poscar_path)
    poscar_structure = Structure.from_file(poscar_path)
    # compare structures to make sure they're the same
    structure_matcher = StructureMatcher(ltol=0.01, stol=0.1, angle_tol=1.0)
    structures_are_equivalent = structure_matcher.fit(
        orig_poscar_structure, poscar_structure
    )
    assert structures_are_equivalent


def test_pcat(tmp_path):
    test_file = tmp_path / "pcat.txt"
    with open(test_file, "w+") as fw:
        fw.write(input_string)
    cat_output = pcat(test_file)
    assert cat_output == input_string


def test_pgrep(tmp_path):
    test_file = tmp_path / "pcat.txt"
    with open(test_file, "w+") as fw:
        fw.write(input_string)
    grep_output = pgrep(test_file, "DAV")
    assert len(grep_output) == 2
    grep_output_as_string = "\n".join([line for line in grep_output])
    matching_lines = "\n".join([line.strip() for line in input_string.split("\n")[:2]])
    assert grep_output_as_string == matching_lines


def test_pgrep_as_string(tmp_path):
    test_file = tmp_path / "pcat.txt"
    with open(test_file, "w+") as fw:
        fw.write(input_string)
    grep_output_as_string = pgrep(test_file, "DAV", as_string=True)
    matching_lines = "\n".join([line.strip() for line in input_string.split("\n")[:2]])
    assert grep_output_as_string == matching_lines


def test_phead(tmp_path):
    test_file = tmp_path / "phead.txt"
    with open(test_file, "w+") as fw:
        fw.write(input_string)
    n_lines = 2
    head_output = phead(test_file, n_head=n_lines)
    assert len(head_output) == 2
    head_output_as_string = "\n".join([line for line in head_output])
    matching_lines = "\n".join(
        [line.strip() for line in input_string.split("\n")[:n_lines]]
    )
    assert head_output_as_string == matching_lines


def test_phead_as_string(tmp_path):
    test_file = tmp_path / "pcat.txt"
    with open(test_file, "w+") as fw:
        fw.write(input_string)
    head_output_as_string = pgrep(test_file, "DAV", as_string=True)
    matching_lines = "\n".join([line.strip() for line in input_string.split("\n")[:2]])
    assert head_output_as_string == matching_lines


def test_ptail(tmp_path):
    test_file = tmp_path / "ptail.txt"
    with open(test_file, "w+") as fw:
        fw.write(input_string)
    n_lines = 2
    tail_output = ptail(test_file, n_tail=n_lines)
    assert len(tail_output) == 2
    tail_output_as_string = "\n".join([line for line in tail_output])
    matching_lines = "\n".join(
        [line.strip() for line in input_string.split("\n")[-n_lines:]]
    )
    assert tail_output_as_string == matching_lines


def test_ptail_as_string(tmp_path):
    test_file = tmp_path / "ptail.txt"
    with open(test_file, "w+") as fw:
        fw.write(input_string)
    n_lines = 2
    tail_output_as_string = ptail(test_file, n_tail=n_lines, as_string=True)
    matching_lines = "\n".join(
        [line.strip() for line in input_string.split("\n")[-n_lines:]]
    )
    assert tail_output_as_string == matching_lines


def test_make_potcar_anonymous(tmp_path):
    potcar_path = importlib_resources.files("vasp_manager").joinpath(
        str(Path("tests") / "POTCARS" / "POTCAR_example")
    )
    new_potcar_path = tmp_path / "POTCAR"
    shutil.copy(potcar_path, new_potcar_path)
    make_potcar_anonymous(new_potcar_path)
    with open(new_potcar_path) as fr:
        anon_potcar_text = fr.read()
    assert anon_potcar_text == "PAW_PBE P 17Jan2003\nPAW_PBE Pb_d 06Sep2000"



import shutil

import importlib_resources
import numpy as np
import pytest

from vasp_manager.analyzer import ElasticAnalyzer

SUPPORTED_ELASTIC_PROPERTIES = [
    "B_Reuss",
    "B_Voigt",
    "B_VRH",
    "G_Reuss",
    "G_Voigt",
    "G_VRH",
    "unstable",
    "elastic_tensor",
    "vl",
    "vt",
    "vs",
]


@pytest.fixture(scope="session")
def stable_material_dir(tmp_path_factory):
    material_dir = (
        importlib_resources.files("vasp_manager") / "tests" / "calculations" / "material"
    )
    new_material_dir = tmp_path_factory.mktemp("material")
    shutil.copytree(
        material_dir,
        new_material_dir,
        dirs_exist_ok=True,
        symlinks=True,
        ignore=shutil.ignore_patterns("elastic_constants.txt"),
    )
    return new_material_dir


@pytest.fixture(scope="session")
def unstable_material_dir(tmp_path_factory):
    material_dir = (
        importlib_resources.files("vasp_manager")
        / "tests"
        / "calculations"
        / "material_spinu"
    )
    new_material_dir = tmp_path_factory.mktemp("material_spinu")
    shutil.copytree(
        material_dir,
        new_material_dir,
        dirs_exist_ok=True,
        symlinks=True,
        ignore=shutil.ignore_patterns("elastic_constants.txt"),
    )
    return new_material_dir


def test_elastic_analyzer_stable(stable_material_dir):
    """
    Make sure elastic is parsed properly for an elastically stable material
    """
    elastic_dir = stable_material_dir / "elastic"
    assert elastic_dir.exists()
    ea = ElasticAnalyzer.from_calc_dir(elastic_dir)
    results = ea.results
    for property in results:
        assert property in SUPPORTED_ELASTIC_PROPERTIES

    assert results["B_Reuss"] == 24.007
    assert results["B_Voigt"] == 24.007
    assert results["B_VRH"] == 24.007
    assert results["G_Reuss"] == 14.286
    assert results["G_Voigt"] == 14.734
    assert results["G_VRH"] == 14.51
    assert results["unstable"] == False
    assert np.array_equal(
        results["elastic_tensor"],
        [
            [47.9984, 12.0109, 12.0109, -0.0, 0.0, 0.0],
            [12.0109, 47.9984, 12.0109, -0.0, 0.0, 0.0],
            [12.0109, 12.0109, 47.9984, -0.0, -0.0, 0.0],
            [-0.0, -0.0, -0.0, 12.5609, 0.0, -0.0],
            [0.0, 0.0, -0.0, 0.0, 12.5609, -0.0],
            [0.0, 0.0, 0.0, -0.0, -0.0, 12.5609],
        ],
    )
    assert results["vl"] == 4.545
    assert results["vt"] == 2.629
    assert results["vs"] == 2.919


def test_elastic_analyzer_unstable(unstable_material_dir):
    """
    Make sure elastic is parsed properly for an elastically unstable material
    """
    elastic_dir = unstable_material_dir / "elastic"
    assert elastic_dir.exists()
    ea = ElasticAnalyzer.from_calc_dir(elastic_dir)
    results = ea.results
    for property in results:
        assert property in SUPPORTED_ELASTIC_PROPERTIES

    assert results["B_Reuss"] == 176.275
    assert results["B_Voigt"] == 176.275
    assert results["B_VRH"] == 176.275
    assert results["G_Reuss"] == -199.401
    assert results["G_Voigt"] == 25.595
    assert results["G_VRH"] == -86.903
    assert results["unstable"] == True
    assert np.array_equal(
        results["elastic_tensor"],
        [
            [140.5505, 194.1377, 194.1377, -0.0, -0.0, 0.0],
            [194.1377, 140.5505, 194.1377, -0.0, -0.0, 0.0],
            [194.1377, 194.1377, 140.5505, -0.0, 0.0, 0.0],
            [-0.0, -0.0, -0.0, 60.5209, 0.0, 0.0],
            [-0.0, -0.0, 0.0, 0.0, 60.5209, -0.0],
            [0.0, 0.0, 0.0, -0.0, -0.0, 60.5209],
        ],
    )
    assert results["vl"] == 2.933
    assert np.isnan(results["vt"])
    assert np.isnan(results["vs"])



import filecmp
import shutil

import importlib_resources
import pytest
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core import Structure

from vasp_manager.calculation_manager import (
    BulkmodCalculationManager,
    ElasticCalculationManager,
    RlxCalculationManager,
    RlxCoarseCalculationManager,
    StaticCalculationManager,
)
from vasp_manager.utils import pgrep

"""
Four material paths exist in calculations/
1) material: NaCl completed successfully
2) material_needs_rerun: NaCl with failed runs to rerun from scratch
3) material_needs_archive: NaCl with failed runs to make archive directories
4) material_hit_erros: NaCl with errors in stdout.txt and stderr.txt
5) material_spinu: CuO with spin and DFT+U completed sucessfully
"""

INPUT_FILES = ["POSCAR", "POTCAR", "INCAR", "vasp.q"]


@pytest.fixture(scope="session")
def calcs_dir(tmp_path_factory):
    """
    Complete calculations directory
    """
    calculations_dir = (
        importlib_resources.files("vasp_manager") / "tests" / "calculations"
    )
    new_calculations_dir = tmp_path_factory.mktemp("calculations")
    shutil.copytree(
        calculations_dir,
        new_calculations_dir,
        dirs_exist_ok=True,
        symlinks=True,
    )
    return new_calculations_dir


@pytest.fixture(scope="session")
def base_calcs_dir(tmp_path_factory):
    """
    Calculations directory that needs each run type to be set up
    """
    calculations_dir = (
        importlib_resources.files("vasp_manager") / "tests" / "calculations"
    )
    new_calculations_dir = tmp_path_factory.mktemp("calculations")
    shutil.copytree(
        calculations_dir,
        new_calculations_dir,
        dirs_exist_ok=True,
        symlinks=True,
        ignore=shutil.ignore_patterns("rlx*", "static", "bulkmod", "elastic"),
    )
    return new_calculations_dir


"""
Do testing of the results first as they don't modify the directories
"""


def test_rlx_coarse_results(calcs_dir):
    """
    Assert rlx-coarse results are parsed properly
    """
    material_dir = calcs_dir / "material"
    rlx_coarse_dir = material_dir / "rlx-coarse"
    assert rlx_coarse_dir.exists()
    rlx_coarse_manager = RlxCoarseCalculationManager(
        material_dir=material_dir,
        to_rerun=True,
        to_submit=False,
    )
    assert rlx_coarse_manager.is_done
    assert rlx_coarse_manager.results == "done"


def test_rlx_results(calcs_dir):
    """
    Assert rlx results are parsed properly
    """
    material_dir = calcs_dir / "material"
    rlx_dir = material_dir / "rlx"
    assert rlx_dir.exists()
    rlx_manager = RlxCalculationManager(
        material_dir=material_dir,
        to_rerun=True,
        to_submit=False,
    )
    assert rlx_manager.is_done
    assert rlx_manager.results["initial_spacegroup"] == 225
    assert rlx_manager.results["relaxed_spacegroup"] == 225
    assert rlx_manager.results["total_dV"] == 0.0379


def test_static_results(calcs_dir):
    """
    Assert static results are parsed properly
    """
    material_dir = calcs_dir / "material"
    static_dir = material_dir / "static"
    assert static_dir.exists()
    static_manager = StaticCalculationManager(
        material_dir=material_dir,
        to_rerun=True,
        to_submit=False,
    )
    assert static_manager.is_done
    assert static_manager.results["final_energy"] == -6.7778924
    assert static_manager.results["final_energy_pa"] == -3.3889462
    assert static_manager.results["magmom_pa"] == None


def test_static_spin_results(calcs_dir):
    """
    Assert static results w/ spin are parsed properly
    """
    material_dir = calcs_dir / "material_spinu"
    static_dir = material_dir / "static"
    assert static_dir.exists()
    static_manager = StaticCalculationManager(
        material_dir=material_dir,
        to_rerun=True,
        to_submit=False,
    )
    assert static_manager.is_done
    assert static_manager.results["final_energy"] == -8.1574727
    assert static_manager.results["final_energy_pa"] == -4.07873635
    assert static_manager.results["magmom_pa"] == 0.0001


def test_bulkmod_results(calcs_dir):
    """
    Assert bulkmod results are parsed properly
    """
    material_dir = calcs_dir / "material"
    bulkmod_dir = material_dir / "bulkmod"
    assert bulkmod_dir.exists()
    bulkmod_manager = BulkmodCalculationManager(
        material_dir=material_dir,
        to_rerun=True,
        to_submit=False,
    )
    assert bulkmod_manager.is_done


def test_elastic_results(calcs_dir):
    """
    Assert elastic results are parsed properly
    """
    material_dir = calcs_dir / "material"
    elastic_dir = material_dir / "elastic"
    assert elastic_dir.exists()
    elastic_manager = ElasticCalculationManager(
        material_dir=material_dir,
        to_rerun=True,
        to_submit=False,
    )
    assert elastic_manager.is_done


"""
Do testing of the errors next as they are independent
"""


def test_hit_errors_and_restart(calcs_dir):
    """
    Test parsing and handling of VASP errors
    """
    material_dir = calcs_dir / "material_hit_errors"
    rlx_coarse_dir = material_dir / "rlx-coarse"
    stdout_path = rlx_coarse_dir / "stdout.txt"
    stderr_path = rlx_coarse_dir / "stderr.txt"
    assert rlx_coarse_dir.exists()
    rlx_coarse_manager = RlxCoarseCalculationManager(
        material_dir=material_dir,
        to_rerun=False,
        to_submit=False,
    )
    errors = rlx_coarse_manager._check_vasp_errors(stdout_path, stderr_path)
    all_errors_addressed = rlx_coarse_manager._address_vasp_errors(errors)
    for error in ["Sub-Space-Matrix", "Inconsistent Bravais", "oom-kill"]:
        assert error in errors
    assert all_errors_addressed
    assert not rlx_coarse_manager.is_done
    assert not rlx_coarse_manager.stopped
    assert rlx_coarse_manager.vasp_input_creator.calc_config["algo"] == "Fast"
    assert rlx_coarse_manager.vasp_input_creator.calc_config["symprec"] == "1e-08"
    assert rlx_coarse_manager.vasp_input_creator.ncore_per_node_for_memory == 64
    assert rlx_coarse_manager.results == "not finished"


def test_hit_errors_and_stop(calcs_dir):
    """
    Test parsing and handling of VASP errors
    """
    material_dir = calcs_dir / "material_hit_errors"
    rlx_dir = material_dir / "rlx"
    stdout_path = rlx_dir / "stdout.txt"
    stderr_path = rlx_dir / "stderr.txt"
    assert rlx_dir.exists()
    rlx_manager = RlxCalculationManager(
        material_dir=material_dir,
        to_rerun=False,
        to_submit=False,
    )
    errors = rlx_manager._check_vasp_errors(stdout_path, stderr_path)
    all_errors_addressed = rlx_manager._address_vasp_errors(errors)
    for error in [
        "num prob",
        "BRMIX",
        "SICK JOB",
        "VERY BAD NEWS",
        "Fatal error",
        "SETYLM",
        "Segmentation",
        "command not found",
    ]:
        assert error in errors
    assert not all_errors_addressed
    assert not rlx_manager.is_done
    assert rlx_manager.stopped
    assert rlx_manager.results == "STOPPED"


def test_rlx_coarse_hit_errors(calcs_dir):
    """
    Test parsing and handling of VASP errors for rlx-coarse
    """
    material_dir = calcs_dir / "material_hit_errors"
    rlx_coarse_dir = material_dir / "rlx-coarse"
    assert rlx_coarse_dir.exists()
    rlx_coarse_manager = RlxCoarseCalculationManager(
        material_dir=material_dir,
        to_rerun=False,
        to_submit=False,
    )
    assert not rlx_coarse_manager.is_done


def test_rlx_hit_errors(calcs_dir):
    """
    Test parsing and handling of VASP errors for rlx
    """
    material_dir = calcs_dir / "material_hit_errors"
    rlx_dir = material_dir / "rlx"
    assert rlx_dir.exists()
    rlx_manager = RlxCalculationManager(
        material_dir=material_dir,
        to_rerun=False,
        to_submit=False,
    )
    assert not rlx_manager.is_done


def test_static_hit_errors(calcs_dir):
    """
    Test parsing and handling of VASP errors for static
    """
    material_dir = calcs_dir / "material_hit_errors"
    static_dir = material_dir / "static"
    assert static_dir.exists()
    static_manager = StaticCalculationManager(
        material_dir=material_dir,
        to_rerun=False,
        to_submit=False,
    )
    assert not static_manager.is_done


def test_bulkmod_hit_errors(calcs_dir):
    """
    Test parsing and handling of VASP errors for bulkmod
    """
    material_dir = calcs_dir / "material_hit_errors"
    bulkmod_dir = material_dir / "bulkmod"
    assert bulkmod_dir.exists()
    bulkmod_manager = BulkmodCalculationManager(
        material_dir=material_dir,
        to_rerun=False,
        to_submit=False,
    )
    assert not bulkmod_manager.is_done


def test_elastic_hit_errors(calcs_dir):
    """
    Test parsing and handling of VASP errors for elastic
    """
    material_dir = calcs_dir / "material_hit_errors"
    elastic_dir = material_dir / "elastic"
    assert elastic_dir.exists()
    elastic_manager = ElasticCalculationManager(
        material_dir=material_dir,
        to_rerun=False,
        to_submit=False,
    )
    assert not elastic_manager.is_done


def test_parse_magmom(calcs_dir):
    """
    Test parsing of magmom
    """
    for material_name, expected_magmom_pa in zip(
        ["material", "material_spinu"], [None, -0.0]
    ):
        material_dir = calcs_dir / material_name
        rlx_dir = material_dir / "rlx"
        assert rlx_dir.exists()
        rlx_manager = RlxCalculationManager(
            material_dir=material_dir,
            to_rerun=True,
            to_submit=False,
        )
        assert rlx_manager._parse_magmom_per_atom() == expected_magmom_pa


"""
Do testing of the reruns/archives last as they modify their directories
"""


def test_too_many_rlx_coarse_archives(calcs_dir):
    """
    Assert rlx-coarse handles too many archives correctly
        (Continues to rlx)
    """
    material_dir = calcs_dir / "material_needs_archive"
    rlx_coarse_dir = material_dir / "rlx-coarse"
    assert rlx_coarse_dir.exists()
    rlx_coarse_manager = RlxCoarseCalculationManager(
        material_dir=material_dir, to_rerun=True, to_submit=False, max_reruns=2
    )
    # archive_0 already exists but allow material to continue to rlx instead
    assert rlx_coarse_manager.is_done


def test_too_many_rlx_archives(calcs_dir):
    """
    Assert rlx handles too many archives correctly
        (Errors out and refuses to continue)
    """
    material_dir = calcs_dir / "material_needs_archive"
    rlx_dir = material_dir / "rlx"
    assert rlx_dir.exists()
    rlx_manager = RlxCalculationManager(
        material_dir=material_dir, to_rerun=True, to_submit=False, max_reruns=2
    )
    # archive_0 already exists
    assert not rlx_manager.is_done


def test_rlx_coarse_archive(calcs_dir):
    """
    Assert unfinished rlx-coarse makes archive and that the symmetrized
        structure for the new POSCAR matches the CONTCAR in the archive
    """
    material_dir = calcs_dir / "material_needs_archive"
    rlx_coarse_dir = material_dir / "rlx-coarse"
    assert rlx_coarse_dir.exists()
    rlx_coarse_manager = RlxCoarseCalculationManager(
        material_dir=material_dir,
        to_rerun=True,
        to_submit=False,
    )
    assert not rlx_coarse_manager.is_done
    assert rlx_coarse_manager.results == "not finished"
    # check archive made correctly
    archive_dir = rlx_coarse_dir / "archive_1"
    assert archive_dir.exists()
    for dir in [rlx_coarse_dir, archive_dir]:
        for f_name in INPUT_FILES:
            assert (dir / f_name).exists()
    archive_dir_contcar = Structure.from_file(archive_dir / "CONTCAR")
    rlx_coarse_poscar = Structure.from_file(rlx_coarse_dir / "POSCAR")
    # compare structures to make sure they're the same
    structure_matcher = StructureMatcher(ltol=0.01, stol=0.1, angle_tol=1.0)
    structures_are_equivalent = structure_matcher.fit(
        archive_dir_contcar, rlx_coarse_poscar
    )
    assert structures_are_equivalent


def test_rlx_archive(calcs_dir):
    """
    Assert unfinished rlx makes archive
    Also ensure rlx that finishes with magmom_pa below magmom_per_atom_cutoff
        restarts without spin
    """
    material_dir = calcs_dir / "material_spinu"
    rlx_dir = material_dir / "rlx"
    assert rlx_dir.exists()
    rlx_manager = RlxCalculationManager(
        material_dir=material_dir,
        to_rerun=True,
        to_submit=False,
        magmom_per_atom_cutoff=1.0,
    )
    assert not rlx_manager.is_done
    assert rlx_manager.results == None
    assert not rlx_manager.vasp_input_creator.use_spin
    # check archive made correctly
    archive_dir = rlx_dir / "archive_0"
    assert archive_dir.exists()
    for dir in [rlx_dir, archive_dir]:
        for f_name in INPUT_FILES:
            assert (dir / f_name).exists()
    incar_path = rlx_dir / "INCAR"
    ispin_line = pgrep(incar_path, "ISPIN", as_string=True)
    ispin_setting = int(ispin_line.split("=")[1].strip())
    assert ispin_setting == 1


def test_static_rerun(calcs_dir):
    """
    Assert static reruns from scratch if failed
    """
    material_dir = calcs_dir / "material_needs_rerun"
    static_dir = material_dir / "static"
    assert static_dir.exists()
    static_manager = StaticCalculationManager(
        material_dir=material_dir,
        to_rerun=True,
        to_submit=False,
    )
    assert not static_manager.is_done
    assert static_manager.results == None
    static_dir_files = list(static_dir.glob("*"))
    assert len(static_dir_files) == 4
    for file in static_dir_files:
        assert file.name in INPUT_FILES


def test_bulkmod_rerun(calcs_dir):
    """
    Assert bulkmod reruns from scratch if failed
    """
    material_dir = calcs_dir / "material_needs_rerun"
    bulkmod_dir = material_dir / "bulkmod"
    assert bulkmod_dir.exists()
    bulkmod_manager = BulkmodCalculationManager(
        material_dir=material_dir,
        to_rerun=True,
        to_submit=False,
    )
    assert not bulkmod_manager.is_done
    assert bulkmod_manager.results == None

    bulkmod_dir_contents = list(bulkmod_dir.glob("*"))
    bulkmod_dir_files = [f for f in bulkmod_dir_contents if f.is_file()]
    strain_dirs = [d for d in bulkmod_dir_contents if d.is_dir()]
    assert len(bulkmod_dir_files) == 4
    for file in bulkmod_dir_files:
        assert file.name in INPUT_FILES
    assert len(strain_dirs) == len(bulkmod_manager.strains)
    for strain_dir in strain_dirs:
        strain_dir_files = list(strain_dir.glob("*"))
        # no vasp.q
        assert len(strain_dir_files) == 3
        for file in strain_dir_files:
            assert file.name in INPUT_FILES


def test_elastic_rerun(calcs_dir):
    """
    Assert elastic reruns from scratch if failed
    """
    material_dir = calcs_dir / "material_needs_rerun"
    elastic_dir = material_dir / "elastic"
    assert elastic_dir.exists()
    elastic_manager = ElasticCalculationManager(
        material_dir=material_dir,
        to_rerun=True,
        to_submit=False,
    )
    assert not elastic_manager.is_done
    assert elastic_manager.results == None
